#include <asm/io.h>
#include <linux/fs.h>
#include <linux/pci.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include "kiri_cmos.h"


/* Global variables */
#if defined(CONFIG_BUFFALO_PLATFORM)
#include <linux/crc7.h>
#include "kiri_common.h"
struct class *kiri_cmos_cls;
struct device *kiri_cmos_dev;
#else
extern struct class *kiri_cls;
struct device *cmos_dev;
#endif
dev_t kirit_cmos;
unsigned long pch_pm_base_addr;
struct pci_dev *pdev = NULL;
unsigned short afterg3_en;


/* ----------------------------------------Wake on lan attribute-------------------------------------- */
static ssize_t wol_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	int rv;
	outb(WOL_OFFSET, CMOS_INDEX_PORT);
	rv = inb(CMOS_DATA_PORT);
	
	if(rv==0){
		return sprintf(buf, "on\n");
	}
	else if(rv==1){
		return sprintf(buf, "off\n");
	}
	return sprintf(buf, "error:%d\n", rv);
}
static ssize_t wol_store(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{

	if((strncmp(buf, "on", strlen("on")) == 0)){
		outb(WOL_OFFSET, CMOS_INDEX_PORT);
		outb(WOL_ENABLE, CMOS_DATA_PORT);
		
		/* set enable by controling register directly */
		outw(inw(pch_pm_base_addr+PM1_STS)|0x4000, pch_pm_base_addr+PM1_STS);
		printk("PM1_STS:%X\n", inw(pch_pm_base_addr+PM1_STS));
		outw(inw(pch_pm_base_addr+PM1_EN)&(~0x4000), pch_pm_base_addr+PM1_EN);
		printk("PM1_EN:%X\n", inw(pch_pm_base_addr+PM1_EN));

	}
	
	if((strncmp(buf, "off", strlen("off")) == 0)){
		outb(WOL_OFFSET, CMOS_INDEX_PORT);
		outb(WOL_DISABLE, CMOS_DATA_PORT);
		
		/* set disable by controling register directly */
		outw(inw(pch_pm_base_addr+PM1_STS)|0x4000, pch_pm_base_addr+PM1_STS);
		printk("PM1_STS:%X\n", inw(pch_pm_base_addr+PM1_STS));
		outw(inw(pch_pm_base_addr+PM1_EN)|0x4000, pch_pm_base_addr+PM1_EN);
		printk("PM1_EN:%X\n", inw(pch_pm_base_addr+PM1_EN));
				
	}

	/* Enable to change CMOS setting */
		outb(CMOS_CHANGE_OFFSET, CMOS_INDEX_PORT);
		outb(CMOS_CHANGE_ENABLE, CMOS_DATA_PORT);

	return count;
}

DEVICE_ATTR(wol_setting, 0644, wol_show, wol_store);

/* ---------------------------------------AC power loss attribute------------------------------------- */
static ssize_t acpower_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	int rv;
	outb(AC_POWER_LOSS_OFFSET, CMOS_INDEX_PORT);
	rv = inb(CMOS_DATA_PORT);
	
	if(rv==0){
		return sprintf(buf, "off\n");
	}
	else if(rv==1){
		return sprintf(buf, "on\n");
	}
	else if(rv==2){
		return sprintf(buf, "last status\n");
	}
	return sprintf(buf, "error:%d", rv);
}
static ssize_t acpower_store(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	int rv;
	
	if((strncmp(buf, "on", strlen("on")) == 0)){
		outb(AC_POWER_LOSS_OFFSET, CMOS_INDEX_PORT);
		outb(AC_POWER_LOSS_POWER_ON, CMOS_DATA_PORT);

		rv = pci_read_config_word(pdev, GEN_PMCON_3, &afterg3_en);
		printk("AFTER_G3:%X \n", afterg3_en);
		afterg3_en &=(~0x0001);
		printk("AFTER_G3:%X \n", afterg3_en);
		rv = pci_write_config_word(pdev, GEN_PMCON_3, afterg3_en);
	}
	
	if((strncmp(buf, "off", strlen("off")) == 0)){
		outb(AC_POWER_LOSS_OFFSET, CMOS_INDEX_PORT);
		outb(AC_POWER_LOSS_POWER_OFF, CMOS_DATA_PORT);
		
		rv = pci_read_config_word(pdev, GEN_PMCON_3, &afterg3_en);
		printk("AFTER_G3:%X \n", afterg3_en);
		afterg3_en |= 0x0001;
		printk("AFTER_G3:%X \n", afterg3_en);
		rv = pci_write_config_word(pdev, GEN_PMCON_3, afterg3_en);
		
	}

	if((strncmp(buf, "last", strlen("last")) == 0)){
		outb(AC_POWER_LOSS_OFFSET, CMOS_INDEX_PORT);
		outb(AC_POWER_LOSS_LAST_STATE, CMOS_DATA_PORT);
		
		rv = pci_read_config_word(pdev, GEN_PMCON_3, &afterg3_en);
		printk("AFTER_G3:%X \n", afterg3_en);
		afterg3_en &= 0xFFFE;
		printk("AFTER_G3:%X \n", afterg3_en);
		rv = pci_write_config_word(pdev, GEN_PMCON_3, afterg3_en);
	}

	/* Enable to change CMOS setting */
		outb(CMOS_CHANGE_OFFSET, CMOS_INDEX_PORT);
		outb(CMOS_CHANGE_ENABLE, CMOS_DATA_PORT);

	return count;
}

DEVICE_ATTR(power_setting, 0644, acpower_show, acpower_store);

/* -----------------------------------------PID value attribute--------------------------------------- */

uint32_t kiri_get_pid(void)
{
	uint32_t ret = 0;
	uint8_t crc_written = 0;
	uint8_t crc_calculated = 0;

	outb(PID_OFFSET_1, CMOS_INDEX_PORT);
	ret |= inb(CMOS_DATA_PORT);
	outb(PID_OFFSET_2, CMOS_INDEX_PORT);
	ret |= (inb(CMOS_DATA_PORT)) << 8;
	outb(PID_OFFSET_3, CMOS_INDEX_PORT);
	ret |= (inb(CMOS_DATA_PORT)) << 16;
	outb(PID_OFFSET_4, CMOS_INDEX_PORT);
	ret |= (inb(CMOS_DATA_PORT)) << 24;

	outb(PID_OFFSET_CRC, CMOS_INDEX_PORT);
	crc_written = inb(CMOS_DATA_PORT);

	crc_calculated = crc7(crc_calculated, (uint8_t *)&ret, 4);

	if(crc_written != crc_calculated)
	{
		printk("WARN! : CRC value is different. crc_written=%02X vs crc_calculated=%02X\n", crc_written, crc_calculated);
		ret = 0;
	}

	return ret;
}
EXPORT_SYMBOL(kiri_get_pid);

static ssize_t pid_show(struct device *dev, struct device_attribute *attr, char *buf)
{
#if defined(CONFIG_BUFFALO_PLATFORM)
	uint32_t ret;
	ret = (unsigned int)kiri_get_pid();
#else
	unsigned int ret;
	int temp1, temp2, temp3, temp4;
	
	outb(PID_OFFSET_1, CMOS_INDEX_PORT);
	temp1 = inb(CMOS_DATA_PORT);
	outb(PID_OFFSET_2, CMOS_INDEX_PORT);
	temp2 = inb(CMOS_DATA_PORT);
	outb(PID_OFFSET_3, CMOS_INDEX_PORT);
	temp3 = inb(CMOS_DATA_PORT);
	outb(PID_OFFSET_4, CMOS_INDEX_PORT);
	temp4 = inb(CMOS_DATA_PORT);
	
	ret = temp4*256*256*256+temp3*256*256+temp2*256+temp1;
#endif

	return sprintf(buf, "0x%08X\n", ret);
}

static ssize_t pid_store(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
#if defined(CONFIG_BUFFALO_PLATFORM)
	uint32_t temp = simple_strtoul(buf,&buf,16);
	uint8_t crc = 0;

	crc = crc7(crc, (uint8_t *)&temp, 4);
#else
	unsigned int temp = simple_strtoul(buf,&buf,16);
#endif
	outb(PID_OFFSET_4, CMOS_INDEX_PORT);
	outb((temp >> 24)&0x000000FF, CMOS_DATA_PORT);
	outb(PID_OFFSET_3, CMOS_INDEX_PORT);
	outb((temp >> 16)&0x000000FF, CMOS_DATA_PORT);
	outb(PID_OFFSET_2, CMOS_INDEX_PORT);
	outb((temp >> 8)&0x000000FF, CMOS_DATA_PORT);
	outb(PID_OFFSET_1, CMOS_INDEX_PORT);
	outb(temp&0x000000FF, CMOS_DATA_PORT);
#if defined(CONFIG_BUFFALO_PLATFORM)
	outb(PID_OFFSET_CRC, CMOS_INDEX_PORT);
	outb(crc, CMOS_DATA_PORT);
#endif
	
	/* Enable to change CMOS setting */
	outb(CMOS_CHANGE_OFFSET, CMOS_INDEX_PORT);
	outb(CMOS_CHANGE_ENABLE, CMOS_DATA_PORT);
	
	return count;
}

DEVICE_ATTR(PID, 0644, pid_show, pid_store);

/* -------------------------------------------Region attribute---------------------------------------- */
static ssize_t region_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	uint8_t region_code = 0;
	uint8_t crc_written = 0;
	uint8_t crc_calculated = 0;

	outb(REGION_OFFSET, CMOS_INDEX_PORT);
	region_code = inb(CMOS_DATA_PORT);

	outb(REGION_OFFSET_CRC, CMOS_INDEX_PORT);
	crc_written = inb(CMOS_DATA_PORT);

	crc_calculated = crc7(crc_calculated, (uint8_t *)&region_code, 1);

	if(crc_written != crc_calculated)
	{
		printk("WARN! : CRC value is different. crc_written=%02X vs crc_calculated=%02X\n", crc_written, crc_calculated);
		region_code = 0;
	}
	
	return sprintf(buf, "0x%02X\n", region_code);
}
static ssize_t region_store(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	unsigned int temp = simple_strtoul(buf,&buf,16);
	uint8_t crc = 0;

	crc = crc7(crc, (uint8_t *)&temp, 1);

	outb(REGION_OFFSET, CMOS_INDEX_PORT);
	outb(temp, CMOS_DATA_PORT);

	outb(REGION_OFFSET_CRC, CMOS_INDEX_PORT);
	outb(crc, CMOS_DATA_PORT);
	
	/* Enable to change CMOS setting */
	outb(CMOS_CHANGE_OFFSET, CMOS_INDEX_PORT);
	outb(CMOS_CHANGE_ENABLE, CMOS_DATA_PORT);
	
	return count;
}

DEVICE_ATTR(region, 0644, region_show, region_store);
/* ------------------------------------------Open the device------------------------------------------ */
int cmos_open(struct inode *inode_ptr, struct file *fptr)
{
	printk("CMOS driver is opened\n");                    
	return(0);
}

/* -----------------------------------------Close the device------------------------------------------ */
int cmos_close(struct inode *inode_ptr, struct file *fptr)
{
	printk("CMOS driver is closed\n");
	return(0);
}

/* -------------------------------IOCTL interface for user to get and set----------------------------- */
int cmos_ioctl(struct file *file, unsigned int cmd, unsigned long para)
{
	// struct inode * inode, 
	unsigned int	write_data=0;
	unsigned char	read_value=0;
	unsigned int	ret = 0;
	int temp1, temp2, temp3, temp4;

	switch(cmd)
	{
		/* Interface to read WOL status in CMOS */
		case GET_WOL:
			outb(WOL_OFFSET, CMOS_INDEX_PORT);
			read_value = inb(CMOS_DATA_PORT);
			printk("Wake on LAN is %x\n",read_value);
			ret = read_value;
			break;

		/* Interface to write WOL setting in CMOS */
		case SET_WOL:
			outb(WOL_OFFSET, CMOS_INDEX_PORT);
			write_data = (unsigned int)para;
			outb(write_data, CMOS_DATA_PORT);
			outb(CMOS_CHANGE_OFFSET, CMOS_INDEX_PORT);
			outb(CMOS_CHANGE_ENABLE, CMOS_DATA_PORT);
			printk("set Wake on LAN to %x\n",write_data);
			break;
			
		/* Interface to read AC Power Loss status in CMOS */
		case GET_ACPOWER:
			outb(AC_POWER_LOSS_OFFSET, CMOS_INDEX_PORT);
			read_value = inb(CMOS_DATA_PORT);
			printk("AC Power Loss is %x\n",read_value);
			ret = read_value;
			break;

		/* Interface to write AC Power Loss setting in CMOS */
		case SET_ACPOWER:
			outb(AC_POWER_LOSS_OFFSET, CMOS_INDEX_PORT);
			write_data = (unsigned int)para;
			outb(write_data, CMOS_DATA_PORT);
			outb(CMOS_CHANGE_OFFSET, CMOS_INDEX_PORT);
			outb(CMOS_CHANGE_ENABLE, CMOS_DATA_PORT);
			printk("set AC Power Loss to %x\n",write_data);
			break;
			
		/* Interface to read PID value in CMOS */
		case GET_PID:
			outb(PID_OFFSET_1, CMOS_INDEX_PORT);
			temp1 = inb(CMOS_DATA_PORT);
			outb(PID_OFFSET_2, CMOS_INDEX_PORT);
			temp2 = inb(CMOS_DATA_PORT);
			outb(PID_OFFSET_3, CMOS_INDEX_PORT);
			temp3 = inb(CMOS_DATA_PORT);
			outb(PID_OFFSET_4, CMOS_INDEX_PORT);
			temp4 = inb(CMOS_DATA_PORT);
			ret = temp4*256*256*256+temp3*256*256+temp2*256+temp1;
			break;

		/* Interface to write PID value in CMOS */
		case SET_PID:
			write_data = (unsigned int)para;
			outb(PID_OFFSET_4, CMOS_INDEX_PORT);
			outb((write_data >> 24)&0x000000FF, CMOS_DATA_PORT);
			outb(PID_OFFSET_3, CMOS_INDEX_PORT);
			outb((write_data >> 16)&0x000000FF, CMOS_DATA_PORT);
			outb(PID_OFFSET_2, CMOS_INDEX_PORT);
			outb((write_data >> 8)&0x000000FF, CMOS_DATA_PORT);
			outb(PID_OFFSET_1, CMOS_INDEX_PORT);
			outb(write_data&0x000000FF, CMOS_DATA_PORT);
	
			/* Enable to change CMOS setting */
			outb(CMOS_CHANGE_OFFSET, CMOS_INDEX_PORT);
			outb(CMOS_CHANGE_ENABLE, CMOS_DATA_PORT);
			break;
			
		/* Interface to read region in CMOS */
		case GET_REGION:
			outb(REGION_OFFSET, CMOS_INDEX_PORT);
			read_value = inb(CMOS_DATA_PORT);
			ret = read_value;
			printk("ret:%x\n", ret);
			break;

		/* Interface to write region in CMOS */
		case SET_REGION:
			write_data = (unsigned int)para;
			outb(REGION_OFFSET, CMOS_INDEX_PORT);
			outb(write_data, CMOS_DATA_PORT);
	
			/* Enable to change CMOS setting */
			outb(CMOS_CHANGE_OFFSET, CMOS_INDEX_PORT);
			outb(CMOS_CHANGE_ENABLE, CMOS_DATA_PORT);
			break;

		default:
			printk("wrong command\n");
			break;
	}		 
	return ret;
}

/* ----------------------------------------Get PCH information---------------------------------------- */
static int pch_setting(void)
{
	int ret;
	
	pdev = pci_get_device(INTEL_PCH_C206_VID, INTEL_PCH_C206_DID, NULL);
	if(!pdev)
		return -ENODEV;
	
	ret = pci_read_config_dword(pdev, PMBASE, &pch_pm_base_addr);
	if(ret)
		return -EINVAL;
		
	pch_pm_base_addr &= 0x0000FFD0;
	
	return 0;
}

/* ----------------------------------File operations for this device---------------------------------- */
struct file_operations cmos_fops = {
open: cmos_open,
unlocked_ioctl: cmos_ioctl,
release: cmos_close,
owner: THIS_MODULE,
};

#if !defined(CONFIG_BUFFALO_PLATFORM)
/* -----------------------------------------Create kiri Class----------------------------------------- */
struct class *kiriClassGet(void)
{
	struct class *tmp_cls;

	if(kiri_cls)
		return kiri_cls;

	tmp_cls = class_create(THIS_MODULE, "kiri");
	if(IS_ERR(tmp_cls))
	{
		printk("%s : class_create failed.", __FUNCTION__);
		return NULL;
	}

	kiri_cls = tmp_cls;
	return kiri_cls;
}
#endif

/* ------------------------------------Initialize and exit module------------------------------------- */
static int __init kiri_cmos_init(void)
{
	int ret;
#if !defined(CONFIG_BUFFALO_PLATFORM)
	struct class *cls;
#endif
	printk("Enter %s module init\n", DEVNAME);

	ret = register_chrdev(kiri_cmos_major, DEVNAME, &cmos_fops);
	if(ret <0)
	{
		printk("%s cannot register\n", DEVNAME);
		return(ret);
	}
	
	/* Create /sys/class/kiri */
#if defined(CONFIG_BUFFALO_PLATFORM)
	kiri_cmos_cls = KiriClassGet();
	if(kiri_cmos_cls)
#else
	cls = kiriClassGet();
	if(kiri_cls)
#endif
	{
		int rv = 0;		
		/* Create /sys/class/kiri/cmos */
#if defined(CONFIG_BUFFALO_PLATFORM)
		kiri_cmos_dev = device_create(kiri_cmos_cls, NULL, kirit_cmos, NULL, "cmos");
#else
		kiri_cmos_dev = device_create(kiri_cls, NULL, kirit_cmos, NULL, "cmos");
#endif
	
		/* Create /sys/class/kiri/coms/WOL */
		rv = device_create_file(kiri_cmos_dev, &dev_attr_wol_setting);
	
		/* Create /sys/class/kiri/cmos/acpower_setting */
		rv = device_create_file(kiri_cmos_dev, &dev_attr_power_setting);
		
		/* Create /sys/class/kiri/coms/PID */
		rv = device_create_file(kiri_cmos_dev, &dev_attr_PID);
 
		/* Create /sys/class/kiri/coms/region */
		rv = device_create_file(kiri_cmos_dev, &dev_attr_region);
	}

	pch_setting();

	printk("kiri_cmos module is initialized\n");
	return 0;
}

static void __exit kiri_cmos_exit(void)
{
	/* Remove /sys/class/kiri/cmos/wol_setting */	
	device_remove_file(kiri_cmos_dev, &dev_attr_wol_setting);
	
	/* Remove /sys/class/kiri/cmos/acpower_setting */
	device_remove_file(kiri_cmos_dev, &dev_attr_power_setting);
	
	/* Remove /sys/class/kiri/cmos/PID */
	device_remove_file(kiri_cmos_dev, &dev_attr_PID);
	
	/* Remove /sys/class/kiri/cmos/region */
	device_remove_file(kiri_cmos_dev, &dev_attr_region);
	
	/* Remove /sys/class/kiri/cmos */
	device_destroy(kiri_cmos_cls, kirit_cmos);

	/* Remove /sys/class/kiri */
#if defined(CONFIG_BUFFALO_PLATFORM)
	KiriClassDestroy();
#else
	//class_destroy(kiri_cls);
#endif
	
	unregister_chrdev(kiri_cmos_major, DEVNAME);
}

module_init(kiri_cmos_init);
module_exit(kiri_cmos_exit);
MODULE_LICENSE("GPL");
