#include <asm/io.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/i2c.h>
#include <linux/mutex.h>
#include "kiri_fan.h"

/* common global variables */
#if defined(CONFIG_BUFFALO_PLATFORM)
#include "kiri_common.h"
struct class *kiri_fan_cls = NULL;
#else
extern struct class *kiri_cls;
#endif
struct device *kiri_fan_dev;
unsigned int EC_BASE_ADDR, IT87_EC_ADDR_REG, IT87_EC_DATA_REG;
dev_t kirit_fan;


/* I2c relative variables */
static unsigned short normal_i2c[] ={0x00, I2C_CLIENT_END};
struct i2c_client *kiri_fan_i2c_client;
static char i2c_temp;
struct mutex update_lock;

/* define 4 steps of fan speed */
static unsigned int fan_threshold = 100;
static unsigned char fan_stop_speed = 0;
static unsigned char fan_slow_speed = 150;
static unsigned char fan_fast_speed = 200;
static unsigned char fan_full_speed = 255;
static unsigned int flag_show_speed_by_string = 1;

struct fan_info_st fan_info[] =
{
	{IT87_EC_FAN_TACO1, IT87_EC_FAN_TACO1_EXT, IT87_EC_FAN_SMART_PWM1, IT87_EC_FAN_PWM1, 0x1},
	{IT87_EC_FAN_TACO2, IT87_EC_FAN_TACO2_EXT, IT87_EC_FAN_SMART_PWM2, IT87_EC_FAN_PWM2, 0x2},
	{IT87_EC_FAN_TACO3, IT87_EC_FAN_TACO3_EXT, IT87_EC_FAN_SMART_PWM3, IT87_EC_FAN_PWM3, 0x4},
};

struct temp_info_st temp_info[] =
{
	{IT87_EC_TMPIN1},
	{IT87_EC_TMPIN2},
	{IT87_EC_TMPIN3},
};


/* ----------------------------------------- kiri_fan i2c driver ------------------------------------- */
static int kiri_fan_detect(struct i2c_client *client, struct i2c_board_info *info)
{

	strlcpy(info->type, "kiri_fan", I2C_NAME_SIZE);
	
	return 0;
}

static int kiri_fan_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	kiri_fan_i2c_client = client;
	mutex_init(&update_lock);	
	return 0;
}

static int kiri_fan_remove(struct i2c_client *client)
{
	normal_i2c[0]= 0x00;
	return 0;
}

static const struct i2c_device_id kiri_fan_id[] =
{
	{ "kiri_fan", 0},
	{ }
};
MODULE_DEVICE_TABLE(i2c, kiri_fan_id);

static struct i2c_driver kiri_fan_driver =
{
		.driver		= {
			.name 	= "kiri_fan",
		},
		.probe		= kiri_fan_probe,
		.remove		= kiri_fan_remove,
		.id_table	= kiri_fan_id,

		.class		= I2C_CLASS_HWMON,
		.detect		= kiri_fan_detect,
		.address_list	= normal_i2c, 
};

/* --------------------------------- fan and temp set/get function ----------------------------------- */
unsigned int set_fan_speed(unsigned char fan_num, unsigned char set_val)
{
	unsigned int	write_data;
	unsigned char	read_value;

	if(fan_info[fan_num].reg_smart_pwm == 0 ||
		fan_info[fan_num].reg_pwm == 0)
	{
		return -EINVAL;
	}

	outb(IT87_EC_FAN_MAIN_CTL, IT87_EC_ADDR_REG);
	read_value = inb(IT87_EC_DATA_REG);
	read_value |= fan_info[fan_num].offset;
	outb(read_value, IT87_EC_DATA_REG);

	write_data = set_val;
	outb(fan_info[fan_num].reg_smart_pwm, IT87_EC_ADDR_REG);
	outb(write_data, IT87_EC_DATA_REG);

	outb(fan_info[fan_num].reg_pwm, IT87_EC_ADDR_REG);
	read_value = inb(IT87_EC_DATA_REG);
	read_value &= 0x7F;
	outb(read_value, IT87_EC_DATA_REG);

	//DPRINTK("set FAN%d speed %d\n", fan_num + 1, write_data);
	return 0;
}

unsigned int get_fan_speed(unsigned char fan_num)
{
	unsigned char	read_value;

	outb(IT87_EC_FAN_MAIN_CTL, IT87_EC_ADDR_REG);
	read_value = inb(IT87_EC_DATA_REG);
	read_value |= fan_info[fan_num].offset;
	outb(read_value, IT87_EC_DATA_REG);

	outb(fan_info[fan_num].reg_smart_pwm, IT87_EC_ADDR_REG);
	read_value = inb(IT87_EC_DATA_REG);

	return read_value;
}

unsigned int get_fan_rpm(unsigned char fan_num)
{
	unsigned int	read_value;
	unsigned int	read_value_ext;
	unsigned int	ret = 0;
	int divisor;
	

	outb(fan_info[fan_num].reg, IT87_EC_ADDR_REG);
	read_value = inb(IT87_EC_DATA_REG);
	outb(fan_info[fan_num].reg_ext, IT87_EC_ADDR_REG);
	read_value_ext = inb(IT87_EC_DATA_REG);
	divisor = (read_value_ext*256+read_value)*2;
	if (divisor > 0) {
		ret = 1350000/divisor;
	}

	if(ret < fan_threshold)
		ret = 0;
	return ret;
}

unsigned int get_temp(unsigned char temp_num)
{
	unsigned char	read_value;
	unsigned int	ret = 0;

	outb(temp_info[temp_num].reg, IT87_EC_ADDR_REG);
	read_value = inb(IT87_EC_DATA_REG);
#if !defined(CONFIG_BUFFALO_PLATFORM)
	printk("TEMP%d is %d\n", temp_num + 1, read_value);
#endif
	ret = read_value;
	return ret;
}

char i2c_get_temp(unsigned short addr)
{
	i2c_temp = 0;
	normal_i2c[0] = addr;
	i2c_add_driver(&kiri_fan_driver);
	mutex_lock(&update_lock);
	i2c_temp = i2c_smbus_read_byte_data(kiri_fan_i2c_client, 0x00);
	mutex_unlock(&update_lock);
	i2c_del_driver(&kiri_fan_driver);
	return i2c_temp;
}

/* ------------------------------------ string and unsigned long converter --------------------------- */
static unsigned long fan_speed_str_to_ul(const char *buf)
{
	if((strncmp(buf, "stop", strlen("stop")) == 0))
		return fan_stop_speed;
	else if((strncmp(buf, "slow", strlen("slow")) == 0))
		return fan_slow_speed;
	else if((strncmp(buf, "fast", strlen("fast")) == 0))
		return fan_fast_speed;
	else if((strncmp(buf, "full", strlen("full")) == 0))
		return fan_full_speed;

	return simple_strtoul(buf, NULL, 10);
}

static unsigned long fan_speed_ul_to_str(const char *buf, size_t len, unsigned long speed)
{
	if(speed <= fan_stop_speed)
		snprintf(buf, len, "stop");
	else if(speed > fan_stop_speed && speed <= fan_slow_speed)
		snprintf(buf, len, "slow");
	else if(speed > fan_slow_speed && speed <= fan_fast_speed)
		snprintf(buf, len, "fast");
	else if(speed > fan_fast_speed)
		snprintf(buf, len, "full");
	else
		snprintf(buf, len, "Unknown");

	return 0;
}

/* ------------------------------------------fan_rpm attribute---------------------------------------- */

static ssize_t show_fan1_rpm(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", get_fan_rpm(0));
}

static ssize_t show_fan2_rpm(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", get_fan_rpm(1));
}

static ssize_t show_fan3_rpm(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", get_fan_rpm(2));
}

DEVICE_ATTR(fan1_rpm, 0644, show_fan1_rpm, NULL);
DEVICE_ATTR(fan2_rpm, 0644, show_fan2_rpm, NULL);
DEVICE_ATTR(fan3_rpm, 0644, show_fan3_rpm, NULL);
/* ------------------------------------------fan_speed attribute---------------------------------------- */

static ssize_t show_fan1_speed(struct device *dev, struct device_attribute *attr, char *buf)
{

	if(flag_show_speed_by_string == 1)
	{
		char szSpeed[32];
		fan_speed_ul_to_str(szSpeed, sizeof(szSpeed), get_fan_speed(0));
		return sprintf(buf, "%s\n", szSpeed);
	}

	return sprintf(buf, "%d\n", get_fan_speed(0));
}
static ssize_t store_fan1_speed(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	unsigned long write_val = fan_speed_str_to_ul((const char*) buf );
	if(write_val > 0xff)
		return -EINVAL;

	set_fan_speed(0, (unsigned char)(write_val & 0xff));
#if !defined(CONFIG_BUFFALO_PLATFORM)
	printk("Write value is %lx\n", write_val);
#endif
	return count;
}

static ssize_t show_fan2_speed(struct device *dev, struct device_attribute *attr, char *buf)
{
	if(flag_show_speed_by_string == 1)
	{
		char szSpeed[32];
		fan_speed_ul_to_str(szSpeed, sizeof(szSpeed), get_fan_speed(1));
		return sprintf(buf, "%s\n", szSpeed);
	}

	return sprintf(buf, "%d\n", get_fan_speed(1));
}
static ssize_t store_fan2_speed(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	unsigned long write_val = fan_speed_str_to_ul((const char*) buf );
	if(write_val > 0xff)
		return -EINVAL;

	set_fan_speed(1, (unsigned char)(write_val & 0xff));
	return count;
}

static ssize_t show_fan3_speed(struct device *dev, struct device_attribute *attr, char *buf)
{
	if(flag_show_speed_by_string == 1)
	{
		char szSpeed[32];
		fan_speed_ul_to_str(szSpeed, sizeof(szSpeed), get_fan_speed(2));
		return sprintf(buf, "%s\n", szSpeed);
	}
	return sprintf(buf, "%d\n", get_fan_speed(2));
}
static ssize_t store_fan3_speed(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	unsigned long write_val = fan_speed_str_to_ul((const char*) buf );
	if(write_val > 0xff)
		return -EINVAL;

	set_fan_speed(2, (unsigned char)(write_val & 0xff));
	return count;
}

static ssize_t show_show_speed_by_string(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", flag_show_speed_by_string);
}

static ssize_t store_show_speed_by_string(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	unsigned long value = simple_strtoul(buf, NULL, 10);
	if(value != 0 && value != 1)
		return -EINVAL;

	flag_show_speed_by_string = value;

	return count;
}

DEVICE_ATTR(fan1_speed, 0644, show_fan1_speed, store_fan1_speed);
DEVICE_ATTR(fan2_speed, 0644, show_fan2_speed, store_fan2_speed);
DEVICE_ATTR(fan3_speed, 0644, show_fan3_speed, store_fan3_speed);
DEVICE_ATTR(show_speed_by_string, 0644, show_show_speed_by_string, store_show_speed_by_string);

/* ------------------------------------ set speed pattern attribute ---------------------------------- */

static ssize_t show_fan_stop_speed(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", fan_stop_speed);
}
static ssize_t store_fan_stop_speed(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	unsigned long tmp_speed = simple_strtoul(buf, NULL, 10);
	if(tmp_speed > 0xff)
		return -EINVAL;

	fan_stop_speed = tmp_speed;
	return count;
}

static ssize_t show_fan_slow_speed(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", fan_slow_speed);
}
static ssize_t store_fan_slow_speed(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	unsigned long tmp_speed = simple_strtoul(buf, NULL, 10);
	if(tmp_speed > 0xff)
		return -EINVAL;

	fan_slow_speed = tmp_speed;
	return count;
}

static ssize_t show_fan_fast_speed(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", fan_fast_speed);
}
static ssize_t store_fan_fast_speed(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	unsigned long tmp_speed = simple_strtoul(buf, NULL, 10);
	if(tmp_speed > 0xff)
		return -EINVAL;

	fan_fast_speed = tmp_speed;
	return count;
}

static ssize_t show_fan_full_speed(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", fan_full_speed);
}
static ssize_t store_fan_full_speed(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	unsigned long tmp_speed = simple_strtoul(buf, NULL, 10);
	if(tmp_speed > 0xff)
		return -EINVAL;

	fan_full_speed = tmp_speed;
	return count;
}

DEVICE_ATTR(fan_stop_speed, 0644, show_fan_stop_speed, store_fan_stop_speed);
DEVICE_ATTR(fan_slow_speed, 0644, show_fan_slow_speed, store_fan_slow_speed);
DEVICE_ATTR(fan_fast_speed, 0644, show_fan_fast_speed, store_fan_fast_speed);
DEVICE_ATTR(fan_full_speed, 0644, show_fan_full_speed, store_fan_full_speed);

/* -------------------------------------------temp_attribute------------------------------------------ */



static ssize_t temp1_show(struct device *dev, struct device_attribute *attr, char *buf)

{
	return sprintf(buf, "%d\n", get_temp(0));
}

static ssize_t temp2_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", get_temp(1));
}

static ssize_t temp3_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", get_temp(2));
}

static ssize_t temp4_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", i2c_get_temp(0x48));
}

static ssize_t temp5_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	return sprintf(buf, "%d\n", i2c_get_temp(0x49));
}

DEVICE_ATTR(temp1, 0444, temp1_show, NULL);
DEVICE_ATTR(temp2, 0444, temp2_show, NULL);
DEVICE_ATTR(temp3, 0444, temp3_show, NULL);
DEVICE_ATTR(temp4, 0444, temp4_show, NULL);
DEVICE_ATTR(temp5, 0444, temp5_show, NULL);

/* -----------------------------------------Open the device------------------------------------------- */
int fan_open(struct inode *inode_ptr, struct file *fptr)
{
	printk("Fan driver is opened\n");                    
	return(0);
}

/* -----------------------------------------Close the device------------------------------------------ */
int fan_close(struct inode *inode_ptr, struct file *fptr)
{
	printk("Fan driver is closed\n");
	return(0);
}

/* -----------------------------IOCTL interface for user to get and set------------------------------- */
int fan_ioctl(struct file *file,unsigned int cmd, unsigned long para)
{
	int	ret = 0;
	unsigned char dev_id = (unsigned char)(cmd & ~IOCTL_CMD_MASK);

	switch(cmd & IOCTL_CMD_MASK)
	{
		case GET_TEMP_CMD_BASE:
			if(dev_id > 2){ 
					if(dev_id==3)
						ret = i2c_get_temp(0x48);
					if(dev_id==4)
						ret = i2c_get_temp(0x49);
			}
			else
				ret = get_temp(dev_id);
			break;
		
		case GET_FAN_CMD_BASE:
			ret = get_fan_rpm(dev_id);
			break;
		
		case SET_FAN_CMD_BASE:
			set_fan_speed(dev_id, (unsigned char )para);
			break;
		
		default:
			break;
	}		 
	return ret;
}

/* ----------------------------------File operations for this device-----------------------------------*/
struct file_operations fan_fops = {
	open: fan_open,
	unlocked_ioctl: fan_ioctl,
	release: fan_close,
	owner: THIS_MODULE,
};

#if !defined(CONFIG_BUFFALO_PLATFORM)
/* -----------------------------------------Create kiri Class----------------------------------------- */
struct class *kiriFanClassGet(void)
{
	struct class *tmp_cls;

	if(kiri_cls)
		return kiri_cls;

	tmp_cls = class_create(THIS_MODULE, "kiri");
	if(IS_ERR(tmp_cls))
	{
		printk("%s : class_create failed.", __FUNCTION__);
		return NULL;
	}

	kiri_cls = tmp_cls;
	return kiri_cls;
}
#endif

/* ----------------------------------------ITE8721 MBPnP Config--------------------------------------- */

void EnterMBPnP(void)
{
	outb(0x87, CONFIG_ADDR_PORT);
	outb(0x01, CONFIG_ADDR_PORT);
	outb(0x55, CONFIG_ADDR_PORT);
	outb(0x55, CONFIG_ADDR_PORT);
}

void SetMBPnP(void)
{
	unsigned short BASE_ADDR_EXT;
	unsigned short BASE_ADDR;	
	outb(IT87_LDN_ADDR, CONFIG_ADDR_PORT);
	outb(IT87_EC_LDN, CONFIG_DATA_PORT);
	outb(0x60,CONFIG_ADDR_PORT);
	BASE_ADDR_EXT = inb(CONFIG_DATA_PORT)*256;
	outb(0x61,CONFIG_ADDR_PORT);
	BASE_ADDR = inb(CONFIG_DATA_PORT);
    	EC_BASE_ADDR = BASE_ADDR_EXT+BASE_ADDR;
    	IT87_EC_ADDR_REG = EC_BASE_ADDR+0x05;
    	IT87_EC_DATA_REG = EC_BASE_ADDR+0x06;
}

void ExitMBPnP(void)
{
	outb(0x02, CONFIG_ADDR_PORT);
	outb(0x02, CONFIG_DATA_PORT);		
}

/* ------------------------------------Initialize and exit module------------------------------------- */

static int __init kiri_fan_init(void)
{  	
	int ret;
#if !defined(CONFIG_BUFFALO_PLATFORM)
	struct class *cls;
#endif
	printk("Enter kiri_fan module init.\n");
		
	ret = register_chrdev(kiri_fan_major, DEVNAME, &fan_fops);	
	if(ret<0)
	{
		printk("%s cannot register\n", DEVNAME);
		return (ret);
	}

	/* ITE8721 Config */
	EnterMBPnP();
	SetMBPnP();
	ExitMBPnP();
	
#if defined(CONFIG_BUFFALO_PLATFORM)
	kiri_fan_cls = KiriClassGet();
	if(kiri_fan_cls)
#else
	cls = kiriFanClassGet();
	if(kiri_cls)
#endif
	{
		int rv = 0;		
		/* Create /sys/class/kiri/fan */
#if defined(CONFIG_BUFFALO_PLATFORM)
		kiri_fan_dev = device_create(kiri_fan_cls, NULL, kirit_fan, NULL, "fan");
#else
		kiri_fan_dev = device_create(kiri_cls, NULL, kirit_fan, NULL, "fan");
#endif
	
		/* Create /sys/class/kiri/fan/fan#_speed */
		rv = device_create_file(kiri_fan_dev, &dev_attr_fan1_speed);
		rv = device_create_file(kiri_fan_dev, &dev_attr_fan2_speed);
		rv = device_create_file(kiri_fan_dev, &dev_attr_fan3_speed);

		/* Create /sys/class/kiri/fan/fan_pattern_speed */
		rv = device_create_file(kiri_fan_dev, &dev_attr_fan_stop_speed);
		rv = device_create_file(kiri_fan_dev, &dev_attr_fan_slow_speed);
		rv = device_create_file(kiri_fan_dev, &dev_attr_fan_fast_speed);
		rv = device_create_file(kiri_fan_dev, &dev_attr_fan_full_speed);

		/* Create /sys/class/kiri/fan/temp# */
		rv = device_create_file(kiri_fan_dev, &dev_attr_temp1);
		rv = device_create_file(kiri_fan_dev, &dev_attr_temp2);
		rv = device_create_file(kiri_fan_dev, &dev_attr_temp3);
		rv = device_create_file(kiri_fan_dev, &dev_attr_temp4);
		rv = device_create_file(kiri_fan_dev, &dev_attr_temp5);

		/* Create /sys/class/kiri/fan/fan# */
		rv = device_create_file(kiri_fan_dev, &dev_attr_fan1_rpm);
		rv = device_create_file(kiri_fan_dev, &dev_attr_fan2_rpm);
		rv = device_create_file(kiri_fan_dev, &dev_attr_fan3_rpm);

		rv = device_create_file(kiri_fan_dev, &dev_attr_show_speed_by_string);
	}
	printk("kiri_fan module is initialized.");
	return 0;
}

static void __exit kiri_fan_exit(void)
{

	/* Remove /sys/class/kiri/fan/fan#_speed */	
	device_remove_file(kiri_fan_dev, &dev_attr_fan1_speed);
	device_remove_file(kiri_fan_dev, &dev_attr_fan2_speed);
	device_remove_file(kiri_fan_dev, &dev_attr_fan3_speed);
	
	/* Remove /sys/class/kiri/fan/fan_pattern_speed */	
	device_remove_file(kiri_fan_dev, &dev_attr_fan_stop_speed);
	device_remove_file(kiri_fan_dev, &dev_attr_fan_slow_speed);
	device_remove_file(kiri_fan_dev, &dev_attr_fan_fast_speed);
	device_remove_file(kiri_fan_dev, &dev_attr_fan_full_speed);

	/* Remove /sys/class/kiri/fan/temp# */
	device_remove_file(kiri_fan_dev, &dev_attr_temp1);
	device_remove_file(kiri_fan_dev, &dev_attr_temp2);
	device_remove_file(kiri_fan_dev, &dev_attr_temp3);
	device_remove_file(kiri_fan_dev, &dev_attr_temp4);
	device_remove_file(kiri_fan_dev, &dev_attr_temp5);

	/* Remove /sys/class/kiri/fan/fan# */	
	device_remove_file(kiri_fan_dev, &dev_attr_fan1_rpm);
	device_remove_file(kiri_fan_dev, &dev_attr_fan2_rpm);
	device_remove_file(kiri_fan_dev, &dev_attr_fan3_rpm);

	device_remove_file(kiri_fan_dev, &dev_attr_show_speed_by_string);

	/* Remove /sys/class/kiri/fan */
	device_destroy(kiri_fan_cls, kirit_fan);

	/* Remove /sys/class/kiri */
#if defined(CONFIG_BUFFALO_PLATFORM)
	KiriClassDestroy();
#endif

  //unregister_chrdev_region(kirit_fan, dev_count);
	unregister_chrdev(kiri_fan_major, DEVNAME);
}

module_init(kiri_fan_init);
module_exit(kiri_fan_exit);
MODULE_LICENSE("GPL");
