#include <asm/io.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/delay.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/gpio.h>
#include <linux/version.h>
#include "kiri_rpsu.h"

#define CONFIG_BUFFALO_USE_INTERRUPT_RPSU_DRIVER

/* common global variables */
#if defined(CONFIG_BUFFALO_PLATFORM)
#include "kiri_common.h"
#include <buffalo/kernevnt.h>
struct class *kiri_rpsu_cls = NULL;
#else
struct class *kiri_cls = NULL;
EXPORT_SYMBOL(kiri_cls);
#endif
struct device *kiri_rpsu;
dev_t kirit_rpsu;
static unsigned int pch_c206_gpio_base_addr;
static unsigned int gp_lvl_data;
#if defined(CONFIG_BUFFALO_USE_INTERRUPT_RPSU_DRIVER)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,39)
spinlock_t kiri_rpsu_status_lock = SPIN_LOCK_UNLOCKED;
#else
spinlock_t kiri_rpsu_status_lock;
#endif

/* APIC rgister access port */
unsigned char *index;
unsigned char *data;
unsigned int value;

/* PSU status flag */
int status = 1;
#else
/* PSU status flag */
unsigned int status = 0;
unsigned int pre_status = 0;
unsigned int change_flag = 0;
unsigned int pre_change_flag = 0;
unsigned int count=10;

struct timer_list psu_alarm_polling_timer;
#endif

/* -------------------------------------------set PSU alarm mute-------------------------------------- */
void set_mute(void)
{
	gp_lvl_data = inl(pch_c206_gpio_base_addr+GP_LVL);
	outl((gp_lvl_data & 0xFFFFFFF7), (pch_c206_gpio_base_addr+GP_LVL));		
	mdelay(100);
	outl((gp_lvl_data | 0x00000008), (pch_c206_gpio_base_addr+GP_LVL));
}

/* --------------------------------------------alarm attribute---------------------------------------- */
static ssize_t alarm_store(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{	
	
	if(strncmp(buf, "mute", strlen("mute"))==0) 
	{
		set_mute();			
		printk("set mute\n");
	}

	return count;
}
DEVICE_ATTR(alarm, 0644, NULL, alarm_store);

/* -------------------------------------------status attribute---------------------------------------- */
static ssize_t status_show(struct device *dev, struct device_attribute *attr, char *buf)
{
	if(status==1)
	{		
		return sprintf(buf, "fine\n");
	}
	else	return sprintf(buf, "broken\n");
}
DEVICE_ATTR(status, 0444, status_show, NULL);

#if defined(CONFIG_BUFFALO_USE_INTERRUPT_RPSU_DRIVER)
/* ---------------------------------Kiri rpsu alarm interrupt handler----------------------------------*/
static irqreturn_t kiri_interrupt(int irq, void *dev_id)
{
	unsigned long flags;
	
#if defined(CONFIG_BUFFALO_PLATFORM)
	char msg[32];
	memset(msg, 0, sizeof(msg));
#endif

	printk("interrupt service\n");
	mdelay(100);
	status = -(status);
	printk("status: %d\n",status);	
	spin_lock_irqsave(&kiri_rpsu_status_lock, flags);
	iowrite8(0x3c, index);
	
	if(status==-1){
		/* mask interrupt */
		value = ioread32(data);
		iowrite32((value|0x00010000), data);	
		
		/* set as active-high */
		value = ioread32(data);
		iowrite32((value&0xFFFFDFFF), data);
		
		/* unmask */	
		value = ioread32(data);
		iowrite32((value&0xFFFEFFFF), data);
		
		sprintf(msg, "psu_broken");
		printk("PSU is %s\n",msg);
	}
	else if(status==1){
		/* mask */
		value = ioread32(data);
		iowrite32((value|0x00010000), data);
		
		/* set as active-low */
		value = ioread32(data);
		iowrite32((value|0x00002000), data);
			
		/* unmask */
		value = ioread32(data);
		iowrite32((value&0xFFFEFFFF), data);
		
		sprintf(msg, "psu_fine");
		printk("PSU is %s.\n", msg);
	}

	spin_unlock_irqrestore(&kiri_rpsu_status_lock, flags);

	#if defined(CONFIG_BUFFALO_PLATFORM)
	if(msg[0] != '\0')
		buffalo_kernevnt_queuein(msg);
	#endif

	return IRQ_HANDLED;
}
#else
/* --------------------------------------- Alarm pin status ------------------------------------------ */
void psu_alarm_pin_status(void)
{
	char msg[32];
	memset(msg, 0, sizeof(msg));
	
	status = (inl(pch_c206_gpio_base_addr+GP_LVL)&0x00000010) >> 4;
	if(pre_status == status)
		count--;
	pre_status = status;
	
	if(count == 0){
		if(status && change_flag==0){
			sprintf(msg, "psu_fine");
			printk("PSU is %s.\n", msg);
			change_flag = 1;
		}
	
		else if((!status) && change_flag==1){
			sprintf(msg, "psu_broken");
			printk("PSU is %s.\n", msg);
			change_flag = 0;
		}
		
		#if defined(CONFIG_BUFFALO_PLATFORM)
			if(msg[0] != '\0')
			buffalo_kernevnt_queuein(msg);
		#endif
		count = 10;	
	}
	
	mod_timer(&psu_alarm_polling_timer, jiffies + PSU_ALARM_POLLING_INTERVAL);
}

/* -------------------------------- Polling function for PSU alarm pin ------------------------------- */
static unsigned long Psu_Alarm_Polling_Start(void)
{
  init_timer(&psu_alarm_polling_timer);
  psu_alarm_polling_timer.expires = jiffies + PSU_ALARM_POLLING_INTERVAL;
  psu_alarm_polling_timer.function = psu_alarm_pin_status;
  add_timer(&psu_alarm_polling_timer);
}
#endif

/* -----------------------------------GPIO and APIC registers setting--------------------------------- */
static int pch_c206_gpio_apic_setting(void)
{
	struct pci_dev *pdev = NULL;
	int ret;
	
	pdev = pci_get_device(INTEL_PCH_C206_VID, INTEL_PCH_C206_DID, NULL);
	if(!pdev)
		return -ENODEV;
	
	/* get GPIOBASE */
	ret = pci_read_config_dword(pdev, GPIOBASE, &pch_c206_gpio_base_addr);
	
	if(ret){
		printk("pci_read_config_dword is failed\n");
		return 0;
	}
	
	pch_c206_gpio_base_addr &= 0x0000FFD0;
	
#if defined(CONFIG_BUFFALO_USE_INTERRUPT_RPSU_DRIVER)
	/* set gpio4 as native function */	
	outl(inl(pch_c206_gpio_base_addr+GP_USE_SEL)&0xFFFFFFEF, pch_c206_gpio_base_addr+GP_USE_SEL);
	
	/* set gpio3 as gpio function */
	outl(inl(pch_c206_gpio_base_addr+GP_USE_SEL)|0x00000008, pch_c206_gpio_base_addr+GP_USE_SEL);

	/* set gpio3 as output */
	outl(inl(pch_c206_gpio_base_addr+GP_IO_SEL)&0xFFFFFFF7, pch_c206_gpio_base_addr+GP_IO_SEL);
	
	/* APIC indirect register setting */
	index = ioremap(APIC_index, 1);
	data = ioremap(APIC_data, 4);
	
	/* set interrupt as level triggered */
	iowrite8(0x3c, index);
	value = ioread32(data);
	iowrite32((value|0x00008000), data);
	
	/* set active-low */
	value = ioread32(data);
	iowrite32((value|0x00002000), data);
#else
	/* set gpio3,4 as gpio function */
	outl(inl(pch_c206_gpio_base_addr+GP_USE_SEL)|0x00000018, pch_c206_gpio_base_addr+GP_USE_SEL);

	/* set gpio3 as output */
	outl(inl(pch_c206_gpio_base_addr+GP_IO_SEL)&0xFFFFFFF7, pch_c206_gpio_base_addr+GP_IO_SEL);
	
	/* set gpio4 as input */
	outl(inl(pch_c206_gpio_base_addr+GP_IO_SEL)|0x00000010, pch_c206_gpio_base_addr+GP_IO_SEL);
	
	/* get inital status of PSU */
	status = inl(pch_c206_gpio_base_addr+GP_LVL)&0x00000010 >> 4;
		
#endif		
	return 0;
}

#if defined(CONFIG_BUFFALO_PLATFORM)
#else
/* -----------------------------------------Create kiri Class----------------------------------------- */
struct class *kiriClassGet(void)
{
	struct class *tmp_cls;

	if(kiri_cls)
		return kiri_cls;

	tmp_cls = class_create(THIS_MODULE, "kiri");
	if(IS_ERR(tmp_cls))
	{
		printk("%s : class_create failed.", __FUNCTION__);
		return NULL;
	}

	kiri_cls = tmp_cls;

	return kiri_cls;
}
#endif

/* ------------------------------------Initialize and exit module------------------------------------- */
static int __init kiri_rpsu_init(void)
{  	
	int ret;
#if defined(CONFIG_BUFFALO_PLATFORM)
#else
	struct class *cls;
#endif
	
	printk("Enter kiri_rpsu module init.\n");
		
	kirit_rpsu = MKDEV(kiri_rpsu_major, kiri_rpsu_minor);
	ret = register_chrdev_region(kirit_rpsu, dev_count, DEVNAME);
	if(ret<0){
		printk("%s cannot register\n", DEVNAME);
		return (ret);
	}

	/* create /sys/class/kiri */
	kiri_rpsu_cls = KiriClassGet();
	if(kiri_rpsu_cls){
		
		int rv = 0;		
		/* create /sys/class/kiri/rpsu */
		kiri_rpsu = device_create(kiri_rpsu_cls, NULL, kirit_rpsu, NULL, "rpsu");
	
		/* create /sys/class/kiri/rpsu/id */
		rv = device_create_file(kiri_rpsu, &dev_attr_alarm);
		rv = device_create_file(kiri_rpsu, &dev_attr_status);
	}
#if defined(CONFIG_BUFFALO_USE_INTERRUPT_RPSU_DRIVER)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39)
        spin_lock_init(&kiri_rpsu_status_lock);
#endif
#endif

	/* find GPIO base adderss, and set registers relatvie to interrupt */
	ret = pch_c206_gpio_apic_setting();
#if defined(CONFIG_BUFFALO_USE_INTERRUPT_RPSU_DRIVER)
	ret = request_irq(22, kiri_interrupt, IRQF_SHARED, DEVNAME, kiri_interrupt);
		
	printk("kiri_rpsu module has been initialized as a interrupt driver.\n");
#else
	/* use a timer to detect alarm pin */
	Psu_Alarm_Polling_Start();

	printk("kiri_rpsu module has been initialized as a polling driver.\n");
#endif

	return 0;
}

static void __exit kiri_rpsu_exit(void)
{
#if defined(CONFIG_BUFFALO_USE_INTERRUPT_RPSU_DRIVER)
	/* release IRQ source */
	disable_irq(22);
	free_irq(22, kiri_interrupt);
	
	/* release remapping memory */
	iounmap(index);
	iounmap(data);
#else
	del_timer(&psu_alarm_polling_timer);
#endif
	
	/* remove /sys/class/kiri/fan/fan# */	
	device_remove_file(kiri_rpsu, &dev_attr_alarm);
	device_remove_file(kiri_rpsu, &dev_attr_status);

	/* remove /sys/class/kiri/fan */
	device_destroy(kiri_rpsu_cls, kirit_rpsu);

	/* remove /sys/class/kiri */
#if defined(CONFIG_BUFFALO_PLATFORM)
	KiriClassDestroy();
#else
	class_destroy(kiri_cls);
#endif
	
	unregister_chrdev_region(kirit_rpsu, dev_count);

}

module_init(kiri_rpsu_init);
module_exit(kiri_rpsu_exit);
MODULE_LICENSE("GPL");
