/******************************************************************************
* LED control driver - v0.4
* 1. Control system info LED with specific pattern.
* 2. Control power LED.
*
* Author: Simon Chang, UG.
*
* Change history:
* v0.1 Initial version.
* v0.2
*   1. Modify LED control by H/W.
*   2. Add HDD err LED control.
*   3. Add ioctl function.
* v0.3
*   1. Add condition to check '\n' suffix of command.
* v0.4
*		1. Move Lederr function to hdd_ctrl.c.
******************************************************************************/
#include <asm/io.h>
#include <linux/pci.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <asm/delay.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/cdev.h>
#include <linux/timer.h>
#include <asm/param.h>
#include <linux/i2c.h>
#include <linux/mutex.h>


#define ALARM_LED_STATIC_OFF	0xf0000000
#define ALARM_LED_STATIC_ON	0xf1000000
#define ALARM_LED_HW_BLINK	0xf2000000
#define ALARM_LED_SW_BLINK_MSK	0xf0000000


/* Define Vendor and Device ID for Intel PCH C206 */
#define INTEL_PCH_C206_VID	0x8086
#define INTEL_PCH_C206_DID	0x1c56
#define GPIO_BAR_OFFSET		0x48
#define GP_LVL				0x0c
#define GPO_BLINK			0x18
#define POWER_LED_BIT		(1 << 27)
#define ALARM_LED_BIT		(1 << 21)
#define POWER_LED_MASK		(~POWER_LED_BIT)
#define ALARM_LED_MASK		(~ALARM_LED_BIT)


#define DEVNO			126			//device number
#define	DEVNAME			"kiri_led"	//device name

int led_major = DEVNO;
char *led_devname = DEVNAME;



#define TIME_1_0			(HZ)			// 1 sec
#define TIME_0_5			(HZ/2)			// 0.5 sec
#define TIME_0_3			((HZ*3)/10)		// 0.3 sec
#define TIME_BLINK_ON			TIME_0_3
#define TIME_BLINK_OFF			TIME_0_3
#define TIME_INTERVAL			TIME_1_0
#define TIME_END			(4 * TIME_1_0)

#define BLINK_TEN_THOUSAND_OFF		0x10
#define BLINK_TEN_THOUSAND_ON 		0x11
#define BLINK_THOUSAND_OFF		0x20
#define BLINK_THOUSAND_ON 		0x21
#define BLINK_HUNDRED_OFF		0x30
#define BLINK_HUNDRED_ON 		0x31
#define BLINK_TEN_OFF			0x40
#define BLINK_TEN_ON			0x41
#define BLINK_ONE_OFF			0x50
#define BLINK_ONE_ON			0x51
#define BLINK_INTERVAL			0x60
#define BLINK_ENCODE_ZERO		0x70
#define BLINK_END			0xff


/* Define for ioctl */
#define IOCTL_GET_ALARM_LED		0x00
#define IOCTL_SET_ALARM_LED		0x01
#define IOCTL_SET_ALARM_LED_SW_BLINK	0x02
#define IOCTL_GET_POWER_LED		0x03
#define IOCTL_SET_POWER_LED		0x04

/*
#define IOCTL_SET_HDD_LEDERR0	0x10
#define IOCTL_SET_HDD_LEDERR1	0x11
#define IOCTL_SET_HDD_LEDERR2	0x12
#define IOCTL_SET_HDD_LEDERR3	0x13
#define IOCTL_SET_HDD_LEDERR4	0x14
#define IOCTL_SET_HDD_LEDERR5	0x15
#define IOCTL_SET_HDD_LEDERR6	0x16
#define IOCTL_SET_HDD_LEDERR7	0x17
#define IOCTL_SET_HDD_LEDERR8	0x18
#define IOCTL_SET_HDD_LEDERR9	0x19
#define IOCTL_SET_HDD_LEDERR10	0x1a
#define IOCTL_SET_HDD_LEDERR11	0x1b

#define IOCTL_GET_HDD_LEDERR0	0x20
#define IOCTL_GET_HDD_LEDERR1	0x21
#define IOCTL_GET_HDD_LEDERR2	0x22
#define IOCTL_GET_HDD_LEDERR3	0x23
#define IOCTL_GET_HDD_LEDERR4	0x24
#define IOCTL_GET_HDD_LEDERR5	0x25
#define IOCTL_GET_HDD_LEDERR6	0x26
#define IOCTL_GET_HDD_LEDERR7	0x27
#define IOCTL_GET_HDD_LEDERR8	0x28
#define IOCTL_GET_HDD_LEDERR9	0x29
#define IOCTL_GET_HDD_LEDERR10	0x2a
#define IOCTL_GET_HDD_LEDERR11	0x2b
*/

/* Define for PCA9505 GPIO expander */
/*
#define GPI_PORT3		0x03
#define GPI_PORT4		0x04
#define GPO_PORT3		0x0b
#define GPO_PORT4		0x0c

#define HDD_ERRLED0_BIT		(1 << 0)
#define HDD_ERRLED1_BIT		(1 << 1)
#define HDD_ERRLED2_BIT		(1 << 2)
#define HDD_ERRLED3_BIT		(1 << 3)
#define HDD_ERRLED4_BIT		(1 << 4)
#define HDD_ERRLED5_BIT		(1 << 5)
#define HDD_ERRLED6_BIT		(1 << 6)
#define HDD_ERRLED7_BIT		(1 << 7)
#define HDD_ERRLED8_BIT		(1 << 0)
#define HDD_ERRLED9_BIT		(1 << 1)
#define HDD_ERRLED10_BIT	(1 << 2)
#define HDD_ERRLED11_BIT	(1 << 3)

#define HDD_ERRLED0_MASK	(~HDD_ERRLED0_BIT)
#define HDD_ERRLED1_MASK	(~HDD_ERRLED1_BIT)
#define HDD_ERRLED2_MASK	(~HDD_ERRLED2_BIT)
#define HDD_ERRLED3_MASK	(~HDD_ERRLED3_BIT)
#define HDD_ERRLED4_MASK	(~HDD_ERRLED4_BIT)
#define HDD_ERRLED5_MASK	(~HDD_ERRLED5_BIT)
#define HDD_ERRLED6_MASK	(~HDD_ERRLED6_BIT)
#define HDD_ERRLED7_MASK	(~HDD_ERRLED7_BIT)
#define HDD_ERRLED8_MASK	(~HDD_ERRLED8_BIT)
#define HDD_ERRLED9_MASK	(~HDD_ERRLED9_BIT)
#define HDD_ERRLED10_MASK	(~HDD_ERRLED10_BIT)
#define HDD_ERRLED11_MASK	(~HDD_ERRLED11_BIT)
*/

/* Define for LED Command */
#define LED_CMD_ERR		0
#define LED_CMD_ON		1
#define LED_CMD_OFF		2
#define LED_CMD_BLINK	3

/*
#define PCA9505_SLAVE_ADDR	0x20
*/

/* Addresses to scan */
/*
static const unsigned short hdd_lederr_i2c[] = { PCA9505_SLAVE_ADDR, I2C_CLIENT_END };
struct i2c_client *kiri_hdd_lederr_client;
struct kiri_hdderr_data {
	struct mutex update_lock;
	u8 gpio_port_data;
};
*/

#if defined(CONFIG_BUFFALO_PLATFORM)
#include "kiri_common.h"
struct class *kiri_led_cls = NULL;
#else
extern struct class *kiri_cls;
#endif
/* EXPORT_SYMBOL(kiri_cls); */
struct device *kiri_led_dev = NULL;
EXPORT_SYMBOL(kiri_led_dev);
static unsigned int pch_c206_gpio_base_addr;

struct alm_led {
	uint32_t pattern;			// the blink pattern under running
	uint32_t action;			// command from user space
	struct timer_list blink_timer;
	unsigned char blink_schedule[81];
	int sche_index;
};
struct alm_led alarm_led;



/*******************************************************************************
* function: find_c206_gpio_base_addr
* argument:
* retvalue:
* description:
*******************************************************************************/
static int find_c206_gpio_base_addr(void)
{
	struct pci_dev *pdev = NULL;

	pdev = pci_get_device(INTEL_PCH_C206_VID, INTEL_PCH_C206_DID, NULL);
	if (!pdev) return -ENODEV;
	pci_read_config_dword(pdev, GPIO_BAR_OFFSET, &pch_c206_gpio_base_addr);
	pch_c206_gpio_base_addr &= 0x0000ff80;

	return 0;
}

/*******************************************************************************
* function: set_pwr_led_on
* argument:
* retvalue:
* description:
*******************************************************************************/
void set_pwr_led_on(void)
{
	unsigned int tmpgpioval;

	tmpgpioval = inl(pch_c206_gpio_base_addr + GP_LVL);
	tmpgpioval |= POWER_LED_BIT;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GP_LVL);

	tmpgpioval = inl(pch_c206_gpio_base_addr + GPO_BLINK);
	tmpgpioval &= POWER_LED_MASK;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GPO_BLINK);
}

/*******************************************************************************
* function: set_pwr_led_off
* argument:
* retvalue:
* description:
*******************************************************************************/
void set_pwr_led_off(void)
{
	unsigned int tmpgpioval;

	tmpgpioval = inl(pch_c206_gpio_base_addr + GP_LVL);
	tmpgpioval &= POWER_LED_MASK;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GP_LVL);

	tmpgpioval = inl(pch_c206_gpio_base_addr + GPO_BLINK);
	tmpgpioval &= POWER_LED_MASK;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GPO_BLINK);
}

/*******************************************************************************
* function: set_pwr_led_hw_blink
* argument:
* retvalue:
* description:
*******************************************************************************/
void set_pwr_led_hw_blink(void)
{
	unsigned int tmpgpioval;

	tmpgpioval = inl(pch_c206_gpio_base_addr + GP_LVL);
	tmpgpioval |= POWER_LED_BIT;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GP_LVL);

	tmpgpioval = inl(pch_c206_gpio_base_addr + GPO_BLINK);
	tmpgpioval |= POWER_LED_BIT;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GPO_BLINK);
}

/*******************************************************************************
* function: set_alarm_led_on
* argument:
* retvalue:
* description:
*******************************************************************************/
void set_alarm_led_on(void)
{
	unsigned int tmpgpioval;

	tmpgpioval = inl(pch_c206_gpio_base_addr + GP_LVL);
	tmpgpioval &= ALARM_LED_MASK;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GP_LVL);

	tmpgpioval = inl(pch_c206_gpio_base_addr + GPO_BLINK);
	tmpgpioval &= ALARM_LED_MASK;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GPO_BLINK);
}

/*******************************************************************************
* function: set_alarm_led_off
* argument:
* retvalue:
* description:
*******************************************************************************/
void set_alarm_led_off(void)
{
	unsigned int tmpgpioval;

	tmpgpioval = inl(pch_c206_gpio_base_addr + GP_LVL);
	tmpgpioval |= ALARM_LED_BIT;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GP_LVL);

	tmpgpioval = inl(pch_c206_gpio_base_addr + GPO_BLINK);
	tmpgpioval &= ALARM_LED_MASK;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GPO_BLINK);
}

/*******************************************************************************
* function: set_alarm_led_hw_blink
* argument:
* retvalue:
* description:
*******************************************************************************/
void set_alarm_led_hw_blink(void)
{
	unsigned int tmpgpioval;

	tmpgpioval = inl(pch_c206_gpio_base_addr + GP_LVL);
	tmpgpioval &= ALARM_LED_MASK;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GP_LVL);

	tmpgpioval = inl(pch_c206_gpio_base_addr + GPO_BLINK);
	tmpgpioval |= ALARM_LED_BIT;
	outl(tmpgpioval, pch_c206_gpio_base_addr + GPO_BLINK);
}


/*******************************************************************************
* function: KiriClassGet
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
struct class *KiriClassGet(void)
{
	struct class *tmp_cls;

	if(kiri_cls)
		return kiri_cls;

	tmp_cls = class_create(THIS_MODULE, "kiri");
	if(IS_ERR(tmp_cls)) {
		printk("%s : class_create failed.", __FUNCTION__);
		return NULL;
	}

	kiri_cls = tmp_cls;
	return kiri_cls;
}
*/

/*******************************************************************************
* function: alarm_led_check_param
* argument:
* retvalue:
* description:
*******************************************************************************/
int alarm_led_check_param(char * str, size_t count)
{
	if ((count == 3) && (str[2] == '\n')) count--; //"on"
	else if ((count == 4) && (str[3] == '\n')) count--; //"off"
	else if ((count == 6) && (str[5] == '\n')) count--; //"blink" or "[0-9]{5}"

	// Check for "on"
	if (count == 2) {
		if(strncmp(str, "on", 2) == 0) {
			alarm_led.action = ALARM_LED_STATIC_ON;
			return 0;
		}
	}
	// Check for "off"
	else if (count == 3) {
		if(strncmp(str, "off", 3) == 0) {
			alarm_led.action = ALARM_LED_STATIC_OFF;
			return 0;
		}
	}
	else if (count == 5) {
		// Check for "blink"
		if(strncmp(str, "blink", 5) == 0) {
			alarm_led.action = ALARM_LED_HW_BLINK;
			return 0;
		}
        	// Check for "10101" ~ "29999"
        	else if (
			(
				(str[0] >= 0x31) && (str[0] <= 0x32) && 
				(str[1] >= 0x30) && (str[1] <= 0x39) && 
				(str[2] >= 0x30) && (str[2] <= 0x39) && 
				(str[3] >= 0x30) && (str[3] <= 0x39) &&
				(str[4] >= 0x30) && (str[4] <= 0x39)
			) &&
			(
				//zero illegal for info/error code as well as for device id
				(!( (str[1] == 0x30) && (str[2] == 0x30))) && 
				(!( (str[3] == 0x30) && (str[4] == 0x30)))
			)
		) {
			alarm_led.action = 
				((str[0] & 0x0f) << 16) + 
				((str[1] & 0x0f) << 12) +
				((str[2] & 0x0f) << 8) +
				((str[3] & 0x0f) << 4) +
				(str[4] & 0x0f);
			return 0;
		}
	}

	return -EINVAL;
}

/*******************************************************************************
* function: timer_function
* argument:
* retvalue:
* description:
*******************************************************************************/
static void timer_function(unsigned long data)
{
	switch (alarm_led.blink_schedule[alarm_led.sche_index]) {
		case BLINK_TEN_THOUSAND_ON:
		case BLINK_THOUSAND_ON:
		case BLINK_HUNDRED_ON:
		case BLINK_TEN_ON:
		case BLINK_ONE_ON:
			mod_timer(&alarm_led.blink_timer, jiffies + TIME_BLINK_ON);
			//printk("  On\n");
			set_alarm_led_on();
			break;

		case BLINK_TEN_THOUSAND_OFF:
		case BLINK_THOUSAND_OFF:
		case BLINK_HUNDRED_OFF:
		case BLINK_TEN_OFF:
		case BLINK_ONE_OFF:
			mod_timer(&alarm_led.blink_timer, jiffies + TIME_BLINK_OFF);
			//printk("  Off\n");
			set_alarm_led_off();
			break;

		case BLINK_ENCODE_ZERO:
			mod_timer(&alarm_led.blink_timer, jiffies + (5 * TIME_BLINK_ON));
			set_alarm_led_on();
			break;

		case BLINK_INTERVAL:
			mod_timer(&alarm_led.blink_timer, jiffies + TIME_INTERVAL);
			break;

		case BLINK_END:
			alarm_led.sche_index = -1;
			mod_timer(&alarm_led.blink_timer, jiffies + TIME_END);
			//printk("=======================\n");
			//printk("  On\n");
			set_alarm_led_off();
			break;

	}
	alarm_led.sche_index++;
}

/*******************************************************************************
* function: alarm_led_sw_blink_start
* argument:
* retvalue:
* description:
*******************************************************************************/
void alarm_led_sw_blink_start(void)
{
	int i, j, ten_thousand_times, thousand_times, hundred_times, ten_times, one_times;

	// prepare blinking schedule
	alarm_led.sche_index = 0;
	ten_thousand_times = (alarm_led.pattern >> 16) & 0x0f;
	thousand_times = (alarm_led.pattern >> 12) & 0x0f;
	hundred_times = (alarm_led.pattern >> 8) & 0x0f;
	ten_times = (alarm_led.pattern >> 4) & 0x0f;
	one_times = alarm_led.pattern & 0x0f;

	for (i = 0, j = 0 ; i < ten_thousand_times ; i++) {
		alarm_led.blink_schedule[j++] = BLINK_TEN_THOUSAND_ON;
		alarm_led.blink_schedule[j++] = BLINK_TEN_THOUSAND_OFF;
	}
	alarm_led.blink_schedule[j++] = BLINK_INTERVAL;

	if (thousand_times == 0) {
		alarm_led.blink_schedule[j++] = BLINK_ENCODE_ZERO;
		alarm_led.blink_schedule[j++] = BLINK_THOUSAND_OFF;
	}
	else {
		for (i = 0 ; i < thousand_times ; i++) {
			alarm_led.blink_schedule[j++] = BLINK_THOUSAND_ON;
			alarm_led.blink_schedule[j++] = BLINK_THOUSAND_OFF;
		}
	}
	alarm_led.blink_schedule[j++] = BLINK_INTERVAL;

	if (hundred_times == 0) {
		alarm_led.blink_schedule[j++] = BLINK_ENCODE_ZERO;
		alarm_led.blink_schedule[j++] = BLINK_HUNDRED_OFF;
	}
	else {
		for (i = 0 ; i < hundred_times ; i++) {
			alarm_led.blink_schedule[j++] = BLINK_HUNDRED_ON;
			alarm_led.blink_schedule[j++] = BLINK_HUNDRED_OFF;
		}
	}
	alarm_led.blink_schedule[j++] = BLINK_INTERVAL;

	if (ten_times == 0) {
		alarm_led.blink_schedule[j++] = BLINK_ENCODE_ZERO;
		alarm_led.blink_schedule[j++] = BLINK_TEN_OFF;
	}
	else {
		for (i = 0 ; i < ten_times ; i++) {
			alarm_led.blink_schedule[j++] = BLINK_TEN_ON;
			alarm_led.blink_schedule[j++] = BLINK_TEN_OFF;
		}
	}
	alarm_led.blink_schedule[j++] = BLINK_INTERVAL;

	if (one_times == 0) {
		alarm_led.blink_schedule[j++] = BLINK_ENCODE_ZERO;
		alarm_led.blink_schedule[j++] = BLINK_ONE_OFF;
	}
	else {
		for (i = 0 ; i < one_times ; i++) {
			alarm_led.blink_schedule[j++] = BLINK_ONE_ON;
			alarm_led.blink_schedule[j++] = BLINK_ONE_OFF;
		}
	}
	alarm_led.blink_schedule[j] = BLINK_END;

	//Setup timer for blinking schedule
	init_timer(&alarm_led.blink_timer);
	alarm_led.blink_timer.expires = jiffies + BLINK_TEN_ON;
	//printk("=======================\n");
	//printk("  On\n");
	set_alarm_led_on();
	alarm_led.blink_timer.data = 0;
	alarm_led.blink_timer.function = timer_function;
	add_timer(&alarm_led.blink_timer);
}

/*******************************************************************************
* function: alarm_led_proc
* argument:
* retvalue:
* description:
*******************************************************************************/
void alarm_led_proc(void)
{
	if ((alarm_led.pattern & ALARM_LED_SW_BLINK_MSK) != ALARM_LED_SW_BLINK_MSK) {
		// stop timer here
		del_timer(&alarm_led.blink_timer);
	}

	alarm_led.pattern = alarm_led.action;
	if ((alarm_led.action & ALARM_LED_SW_BLINK_MSK) == ALARM_LED_SW_BLINK_MSK) {
		// set led to on or off or hw blink
		switch(alarm_led.action) {
			case ALARM_LED_STATIC_ON:
				set_alarm_led_on();
			break;

			case ALARM_LED_STATIC_OFF:
				set_alarm_led_off();
			break;

			case ALARM_LED_HW_BLINK:
				set_alarm_led_hw_blink();
			break;
		}
	}
	else {
		// set led to sw blink
		printk("sw blink start...\n");
		alarm_led_sw_blink_start();
	}
}

/*******************************************************************************
* function: set_hdd_lederr_on
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
void set_hdd_lederr_on(int hddno)
{
	struct kiri_hdderr_data *gpio_data = i2c_get_clientdata(kiri_hdd_lederr_client);
	u8 temp = 0x01;
	u8 port;

	if (hddno < 8) {
		temp <<= hddno;
		port = GPO_PORT3;
	}
	else if (hddno < 12) {
		hddno -= 8;
		temp <<= hddno;
		port = GPO_PORT4;
	}
	else return;

	mutex_lock(&gpio_data->update_lock);
	gpio_data->gpio_port_data = (u8)i2c_smbus_read_byte_data(kiri_hdd_lederr_client, port);
	gpio_data->gpio_port_data &= ~temp;
	i2c_smbus_write_byte_data(kiri_hdd_lederr_client, port, gpio_data->gpio_port_data);
	mutex_unlock(&gpio_data->update_lock);
}
*/
/*******************************************************************************
* function: set_hdd_lederr_off
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
void set_hdd_lederr_off(int hddno)
{
	struct kiri_hdderr_data *gpio_data = i2c_get_clientdata(kiri_hdd_lederr_client);
	u8 temp = 0x01;
	u8 port;

	if (hddno < 8) {
		temp <<= hddno;
		port = GPO_PORT3;
	}
	else if (hddno < 12) {
		hddno -= 8;
		temp <<= hddno;
		port = GPO_PORT4;
	}
	else return;

	mutex_lock(&gpio_data->update_lock);
	gpio_data->gpio_port_data = (u8)i2c_smbus_read_byte_data(kiri_hdd_lederr_client, port);
	gpio_data->gpio_port_data |= temp;
	i2c_smbus_write_byte_data(kiri_hdd_lederr_client, port, gpio_data->gpio_port_data);
	mutex_unlock(&gpio_data->update_lock);
}
*/

/*******************************************************************************
* function: get_hdd_lederr_status
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
int get_hdd_lederr_status(int hddno)
{
	struct kiri_hdderr_data *gpio_data = i2c_get_clientdata(kiri_hdd_lederr_client);
	u8 temp = 0x01;
	u8 port;

	if (hddno < 8) {
		temp <<= hddno;
		port = GPO_PORT3;
	}
	else if (hddno < 12) {
		hddno -= 8;
		temp <<= hddno;
		port = GPO_PORT4;
	}
	else return -EINVAL;

	mutex_lock(&gpio_data->update_lock);
	gpio_data->gpio_port_data = (u8)i2c_smbus_read_byte_data(kiri_hdd_lederr_client, port);
	mutex_unlock(&gpio_data->update_lock);

	return (gpio_data->gpio_port_data & temp);
}
*/

/*******************************************************************************
* function: check_led_cmd_param
* argument:
* retvalue:
* description:
*******************************************************************************/
static int check_led_cmd_param(char *buf, size_t count)
{
	if (strncmp(buf, "on", 2) == 0) {
		if (count == 2) {
			return LED_CMD_ON;
		}
		else if (count == 3) {
			if (buf[2] == '\n') return LED_CMD_ON;
			else return LED_CMD_ERR;
		}
		else return LED_CMD_ERR;
	}
	else if (strncmp(buf, "off", 3) == 0) {
		if (count == 3) {
			return LED_CMD_OFF;
		}
		else if (count == 4) {
			if (buf[3] == '\n') return LED_CMD_OFF;
			else return LED_CMD_ERR;
		}
		else return LED_CMD_ERR;
	}
	else if (strncmp(buf, "blink", 5) == 0) {
		if (count == 5) {
			return LED_CMD_BLINK;
		}
		else if (count == 6) {
			if (buf[5] == '\n') return LED_CMD_BLINK;
			else return LED_CMD_ERR;
		}
		else return LED_CMD_ERR;
	}
	else return LED_CMD_ERR;

}

/*******************************************************************************
* function: show_alarm_led_pattern
* function: store_alarm_led_pattern
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_alarm_led_pattern(struct device *dev, struct device_attribute *attr, char *buf)
{
	if ((alarm_led.pattern & ALARM_LED_SW_BLINK_MSK) == ALARM_LED_SW_BLINK_MSK) {
		switch (alarm_led.pattern) {
			case ALARM_LED_STATIC_ON:	return sprintf(buf, "on\n");	break;
			case ALARM_LED_STATIC_OFF:	return sprintf(buf, "off\n");	break;
			case ALARM_LED_HW_BLINK:	return sprintf(buf, "blink\n");	break;
		}
	}
	else {
		return sprintf(buf, "%08x\n", alarm_led.pattern);
	}
	return 0;
}
static ssize_t store_alarm_led_pattern(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	int retval;

	retval = alarm_led_check_param(buf, count);
	if (retval) return retval;
	//printk("New alert led command: 0x%08x\n", alarm_led.action);
	alarm_led_proc();

	return count;
}
DEVICE_ATTR(Alarm, 0644, show_alarm_led_pattern, store_alarm_led_pattern);

/*******************************************************************************
* function: show_power_led
* function: store_power_led
* argument:
* retvalue:
* description:
*******************************************************************************/
static ssize_t show_power_led(struct device *dev, struct device_attribute *attr, char *buf)
{
	unsigned int tmpgpioval;

	tmpgpioval = inl(pch_c206_gpio_base_addr + GPO_BLINK);
	if (tmpgpioval & POWER_LED_BIT)
		return sprintf(buf, "blink\n");

	tmpgpioval = inl(pch_c206_gpio_base_addr + GP_LVL);
	if (tmpgpioval & POWER_LED_BIT)
		return sprintf(buf, "on\n");
	else
		return sprintf(buf, "off\n");
}
static ssize_t store_power_led(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_pwr_led_on();		break;
		case LED_CMD_OFF:	set_pwr_led_off();		break;
		case LED_CMD_BLINK:	set_pwr_led_hw_blink();	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Power, 0644, show_power_led, store_power_led);

/*******************************************************************************
* function: show_Lederr0
* function: store_Lederr0
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr0(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(0))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr0(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(0);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(0);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr0, 0644, show_Lederr0, store_Lederr0);
*/

/*******************************************************************************
* function: show_Lederr1
* function: store_Lederr1
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr1(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(1))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr1(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(1);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(1);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr1, 0644, show_Lederr1, store_Lederr1);
*/

/*******************************************************************************
* function: show_Lederr2
* function: store_Lederr2
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr2(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(2))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr2(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(2);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(2);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr2, 0644, show_Lederr2, store_Lederr2);
*/

/*******************************************************************************
* function: show_Lederr3
* function: store_Lederr3
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr3(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(3))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr3(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(3);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(3);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr3, 0644, show_Lederr3, store_Lederr3);
*/

/*******************************************************************************
* function: show_Lederr4
* function: store_Lederr4
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr4(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(4))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr4(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(4);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(4);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr4, 0644, show_Lederr4, store_Lederr4);
*/

/*******************************************************************************
* function: show_Lederr5
* function: store_Lederr5
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr5(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(5))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr5(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(5);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(5);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr5, 0644, show_Lederr5, store_Lederr5);
*/

/*******************************************************************************
* function: show_Lederr6
* function: store_Lederr6
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr6(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(6))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr6(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(6);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(6);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr6, 0644, show_Lederr6, store_Lederr6);
*/

/*******************************************************************************
* function: show_Lederr7
* function: store_Lederr7
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr7(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(7))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr7(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(7);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(7);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr7, 0644, show_Lederr7, store_Lederr7);
*/

/*******************************************************************************
* function: show_Lederr8
* function: store_Lederr8
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr8(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(8))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr8(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(8);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(8);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr8, 0644, show_Lederr8, store_Lederr8);
*/

/*******************************************************************************
* function: show_Lederr9
* function: store_Lederr9
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr9(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(9))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr9(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(9);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(9);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr9, 0644, show_Lederr9, store_Lederr9);
*/

/*******************************************************************************
* function: show_Lederr10
* function: store_Lederr10
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr10(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(10))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr10(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(10);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(10);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr10, 0644, show_Lederr10, store_Lederr10);
*/

/*******************************************************************************
* function: show_Lederr11
* function: store_Lederr11
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static ssize_t show_Lederr11(struct device *dev, struct device_attribute *attr, char *buf)
{
	if (get_hdd_lederr_status(11))
		return sprintf(buf, "off\n");
	else
		return sprintf(buf, "on\n");
}
static ssize_t store_Lederr11(struct device *dev, struct device_attribute *attr, char *buf, size_t count)
{
	switch (check_led_cmd_param(buf, count)) {
		case LED_CMD_ERR:	return -EINVAL;			break;
		case LED_CMD_ON:	set_hdd_lederr_on(11);	break;
		case LED_CMD_OFF:	set_hdd_lederr_off(11);	break;
		default: 			return -EINVAL;			break;
	}
	return count;
}
DEVICE_ATTR(Lederr11, 0644, show_Lederr11, store_Lederr11);
*/

/*******************************************************************************
* function: led_ioctl
* argument:
* retvalue:
* description:
*******************************************************************************/
int led_ioctl(struct file *file,unsigned int cmd, unsigned long para)
{
	int ret = 0;

	switch (cmd) {
		case IOCTL_GET_ALARM_LED:
			if ((alarm_led.pattern & ALARM_LED_SW_BLINK_MSK) == ALARM_LED_SW_BLINK_MSK) {
				switch (para) {
					case ALARM_LED_STATIC_ON:	printk("on\n");	break;
					case ALARM_LED_STATIC_OFF:	printk("off\n");	break;
					case ALARM_LED_HW_BLINK:	printk("blink\n");	break;
				}
			}
			else {
				printk("%08x\n", alarm_led.pattern);
			}
		break;

		case IOCTL_SET_ALARM_LED_SW_BLINK:
			alarm_led.action = (uint32_t)para;
			alarm_led_proc();
		break;

		case IOCTL_SET_ALARM_LED:
			switch (para) {
				case 0:	alarm_led.action = ALARM_LED_STATIC_OFF;	break;
				case 1:	alarm_led.action = ALARM_LED_STATIC_ON;		break;
				case 2:	alarm_led.action = ALARM_LED_HW_BLINK;	break;
			}
			alarm_led_proc();
		break;

		case IOCTL_GET_POWER_LED:
			if (inl(pch_c206_gpio_base_addr + GP_LVL) & POWER_LED_BIT) {
				if (inl(pch_c206_gpio_base_addr + GPO_BLINK) & POWER_LED_BIT) printk("blink\n");
				else printk("on\n");
			}
			else printk("off\n");
		break;

		case IOCTL_SET_POWER_LED:
			switch (para) {
				case 0:	set_pwr_led_off();		break;
				case 1:	set_pwr_led_on();		break;
				case 2:	set_pwr_led_hw_blink();	break;
			}
		break;
/*
		case IOCTL_SET_HDD_LEDERR0:
		case IOCTL_SET_HDD_LEDERR1:
		case IOCTL_SET_HDD_LEDERR2:
		case IOCTL_SET_HDD_LEDERR3:
		case IOCTL_SET_HDD_LEDERR4:
		case IOCTL_SET_HDD_LEDERR5:
		case IOCTL_SET_HDD_LEDERR6:
		case IOCTL_SET_HDD_LEDERR7:
		case IOCTL_SET_HDD_LEDERR8:
		case IOCTL_SET_HDD_LEDERR9:
		case IOCTL_SET_HDD_LEDERR10:
		case IOCTL_SET_HDD_LEDERR11:
			switch (para) {
				case 0:	set_hdd_lederr_off(cmd-IOCTL_SET_HDD_LEDERR0);		break;
				case 1:	set_hdd_lederr_on(cmd-IOCTL_SET_HDD_LEDERR0);		break;
			}
		break;

		case IOCTL_GET_HDD_LEDERR0:
		case IOCTL_GET_HDD_LEDERR1:
		case IOCTL_GET_HDD_LEDERR2:
		case IOCTL_GET_HDD_LEDERR3:
		case IOCTL_GET_HDD_LEDERR4:
		case IOCTL_GET_HDD_LEDERR5:
		case IOCTL_GET_HDD_LEDERR6:
		case IOCTL_GET_HDD_LEDERR7:
		case IOCTL_GET_HDD_LEDERR8:
		case IOCTL_GET_HDD_LEDERR9:
		case IOCTL_GET_HDD_LEDERR10:
		case IOCTL_GET_HDD_LEDERR11:
			if (get_hdd_lederr_status(cmd-IOCTL_GET_HDD_LEDERR0)) {
				ret = 0;
				printk("off\n");
			}
			else {
				ret = 1;
				printk("on\n");
			}
		break;
*/
		default:
			goto IOCTL_WRONG_CMD;
		break;
	}
	return ret;

IOCTL_WRONG_CMD:
	printk("wrong command\n");
	return ret;
}

struct file_operations led_fops = {
	//open: led_open,
	unlocked_ioctl: led_ioctl,
	//release: led_close,
	owner: THIS_MODULE,
};

/*******************************************************************************
* function: kiri_hdd_lederr_detect
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static int kiri_hdd_lederr_detect(struct i2c_client *client, struct i2c_board_info *info)
{
	strlcpy(info->type, "kiri_hdd_ederr", I2C_NAME_SIZE);

	return 0;
}
*/

/*******************************************************************************
* function: kiri_hdd_lederr_probe
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static int kiri_hdd_lederr_probe(struct i2c_client *client, const struct i2c_device_id *id)
{
	struct kiri_hdderr_data *data;

	if (!(data = kzalloc(sizeof(struct kiri_hdderr_data), GFP_KERNEL))) {
		return -ENOMEM;
	}

	kiri_hdd_lederr_client = client;
	data->gpio_port_data = 0;
	i2c_set_clientdata(kiri_hdd_lederr_client, data);
	mutex_init(&data->update_lock);

	// for init PCA9505, because BIOS is not ready.
	struct kiri_hdderr_data *gpio_data = i2c_get_clientdata(kiri_hdd_lederr_client);
	mutex_lock(&gpio_data->update_lock);
	i2c_smbus_write_byte_data(kiri_hdd_lederr_client, 0x1b, 0x00);
	i2c_smbus_write_byte_data(kiri_hdd_lederr_client, 0x1c, 0x00);
	i2c_smbus_write_byte_data(kiri_hdd_lederr_client, 0x0b, 0xff);
	i2c_smbus_write_byte_data(kiri_hdd_lederr_client, 0x0c, 0xff);
	mutex_unlock(&gpio_data->update_lock);



	return 0;
}
*/

/*******************************************************************************
* function: kiri_hdd_lederr_remove
* argument:
* retvalue:
* description:
*******************************************************************************/
/*
static int kiri_hdd_lederr_remove(struct i2c_client *client)
{
	kfree(i2c_get_clientdata(client));
	return 0;
}

static const struct i2c_device_id kiri_hdd_lederr_id[] = {
	{ "kiri_hdd_ederr", 0 },
	{ }
};

static struct i2c_driver kiri_hdd_lederr_driver = {
	.driver = {
		.name	= "kiri_hdd_ederr",
	},
	.probe		= kiri_hdd_lederr_probe,
	.remove		= kiri_hdd_lederr_remove,
	.id_table	= kiri_hdd_lederr_id,
	.class		= I2C_CLASS_SPD,
	.detect		= kiri_hdd_lederr_detect,
	.address_list	= hdd_lederr_i2c,
};
*/

/*******************************************************************************
* function: kiri_led_ctrl_init
* argument:
* retvalue:
* description:
*******************************************************************************/
static int __init kiri_led_ctrl_init(void)
{
	int ret;

	printk(KERN_ALERT "led control module init...\n");
	
/*
	if (i2c_add_driver(&kiri_hdd_lederr_driver)) return -ENODEV;
*/

	if (find_c206_gpio_base_addr()) return -ENODEV;

	/* Register device, and may be allocating major# */
	ret = register_chrdev(led_major, led_devname, &led_fops);
	if (ret < 0) {
		printk("%s cannot register\n", led_devname);
		return(ret);
	}
	else {
		if (led_major == 0) led_major = ret;
	}

	kiri_led_cls = KiriClassGet();
	if(kiri_led_cls){
		if(kiri_led_dev == NULL){
			kiri_led_dev = device_create(kiri_led_cls, NULL, MKDEV(led_major, 0), NULL, "led");
		}
		ret = device_create_file(kiri_led_dev, &dev_attr_Alarm);
		ret = device_create_file(kiri_led_dev, &dev_attr_Power);
/*
		ret = device_create_file(dev, &dev_attr_Lederr0);
		ret = device_create_file(dev, &dev_attr_Lederr1);
		ret = device_create_file(dev, &dev_attr_Lederr2);
		ret = device_create_file(dev, &dev_attr_Lederr3);
		ret = device_create_file(dev, &dev_attr_Lederr4);
		ret = device_create_file(dev, &dev_attr_Lederr5);
		ret = device_create_file(dev, &dev_attr_Lederr6);
		ret = device_create_file(dev, &dev_attr_Lederr7);
		ret = device_create_file(dev, &dev_attr_Lederr8);
		ret = device_create_file(dev, &dev_attr_Lederr9);
		ret = device_create_file(dev, &dev_attr_Lederr10);
		ret = device_create_file(dev, &dev_attr_Lederr11);
*/
	}
	alarm_led.pattern = ALARM_LED_STATIC_OFF;
	alarm_led.action = 0;

	return 0;
}

/*******************************************************************************
* function: kiri_led_ctrl_exit_cleanup
* argument:
* retvalue:
* description:
*******************************************************************************/
static void __exit kiri_led_ctrl_exit_cleanup(void)
{
	//int ret;

	printk(KERN_ALERT "led control module exit...\n");

	device_remove_file(kiri_led_dev, &dev_attr_Alarm);
	device_remove_file(kiri_led_dev, &dev_attr_Power);
	device_destroy(kiri_led_cls, MKDEV(led_major, 0));
	KiriClassDestroy();
	unregister_chrdev(led_major, led_devname);
	/*
	i2c_del_driver(&kiri_hdd_lederr_driver);
	*/
}

module_init(kiri_led_ctrl_init);
module_exit(kiri_led_ctrl_exit_cleanup);
MODULE_LICENSE("GPL");
